import { useSyncExternalStore } from 'react';
import ReactDOM from 'react-dom';
import { Callback, State, Store } from './type';

const isDev = process.env.NODE_ENV === 'development';
const isObject = (x: any) => Object.prototype.toString.call(x) === '[object Object]';

function reox<T extends State>(state: T): T {
  if (isDev && !isObject(state)) throw new Error('object required');

  const _store: Store<T> = {} as Store<T>;

  Object.keys(state).forEach((key: keyof T) => {
    if (typeof state[key] === 'function') return;

    const _listeners = new Set<Callback>();

    _store[key] = {
      subscribe: (listener) => {
        _listeners.add(listener);
        return () => _listeners.delete(listener);
      },
      getSnapshot: () => state[key],
      setSnapshot: (val) => {
        if (val === state[key]) return;
        state[key] = val;
        ReactDOM.unstable_batchedUpdates?.(() => _listeners.forEach((listener) => listener()));
      },
      useSnapshot: () => {
        return useSyncExternalStore(_store[key].subscribe, _store[key].getSnapshot);
      },
    };
  });

  return new Proxy(state, {
    get: (_, key: keyof T) => {
      try {
        return _store[key].useSnapshot();
      } catch (e) {
        return state[key];
      }
    },
    set: (_, key: keyof T, val: T[keyof T]) => {
      _store[key].setSnapshot(val);
      return true;
    },
  } as ProxyHandler<T>);
}

export default reox;